#include <cstring>
#include "matrix.hh"
#include "xyz.hh"

using std::memset;

Matrix::Matrix()
{
    Zero();
}

Matrix::Matrix(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
{
    InitRotate(ax, ay, az);
}

Matrix::Matrix(const XYZ &axis, FLOAT_ARG ang)
{
    InitAxisRotate(axis, ang);
}

void Matrix::Zero()
{
    memset(&val, 0, sizeof val);
}

void Matrix::Identity()
{
    for(unsigned i=0; i<4; ++i)
        for(unsigned j=0; j<4; ++j)
            val[i][j] = (j==i) ? 1.0 : 0.0;
}

void Matrix::InitRotate(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
{
    FLOAT tmp[3][4][4];
    
    Identity();

    memset(&tmp, 0, sizeof(tmp));

    tmp[0][1][1]=tmp[0][2][2]=COS(ax); tmp[0][2][1]=-(tmp[0][1][2]=SIN(ax));
    tmp[1][0][0]=tmp[1][2][2]=COS(ay); tmp[1][0][2]=-(tmp[1][2][0]=SIN(ay));
    tmp[2][0][0]=tmp[2][1][1]=COS(az); tmp[2][1][0]=-(tmp[2][0][1]=SIN(az));

    tmp[0][0][0] = 1;
    tmp[0][3][3] = 1;
    tmp[1][1][1] = 1;
    tmp[1][3][3] = 1;
    tmp[2][2][2] = 1;
    tmp[2][3][3] = 1;

    for(int i=0; i<4; i++)
    {
        FLOAT m2[4];

        for(int j=0; j<4; j++)
            m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                    tmp[1][i][1] * tmp[0][1][j] +
                    tmp[1][i][2] * tmp[0][2][j] +
                    tmp[1][i][3] * tmp[0][3][j];

        for(int j=0; j<4; j++)
            val[i][j] = m2[0] * tmp[2][0][j] +
                        m2[1] * tmp[2][1][j] +
                        m2[2] * tmp[2][2][j] +
                        m2[3] * tmp[2][3][j];
    }
}

/* Math borrowed from Povray, matrices.c:Compute_Axis_Rotation_Transform */
void Matrix::InitAxisRotate(const XYZ &V, FLOAT_ARG ang)
{
    XYZ tmp = V.Normalized();
    Identity();

    const FLOAT cosx = COS(ang);
    const FLOAT sinx = SIN(ang);
    
    val[0][0] = tmp.x * tmp.x + cosx * (1.0 - tmp.x * tmp.x);
    val[0][1] = tmp.x * tmp.y * (1.0 - cosx) + tmp.z * sinx;
    val[0][2] = tmp.x * tmp.z * (1.0 - cosx) - tmp.y * sinx;
    
    val[1][0] = tmp.x * tmp.y * (1.0 - cosx) - tmp.z * sinx;
    val[1][1] = tmp.y * tmp.y + cosx * (1.0 - tmp.y * tmp.y);
    val[1][2] = tmp.y * tmp.z * (1.0 - cosx) + tmp.x * sinx;
    
    val[2][0] = tmp.x * tmp.z * (1.0 - cosx) + tmp.y * sinx;
    val[2][1] = tmp.y * tmp.z * (1.0 - cosx) - tmp.x * sinx;
    val[2][2] = tmp.z * tmp.z + cosx * (1.0 - tmp.z * tmp.z);
}

void Matrix::Rotate(XYZ &p) const
{
    register FLOAT tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
    register FLOAT ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];
    p.z               = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
    p.y = ty;
    p.x = tx;
}

const Matrix Matrix::operator * (const Matrix &m) const
{
    Matrix res;
    for(unsigned i=0; i<3; ++i)
        for(unsigned j=0; j<4; ++j)
            for(unsigned k=0; k<4; ++k)
                res.val[i][j] += val[i][k] * m.val[k][j];
    res.val[3][3] = 1;
    return res;
}
